@@ -19,7 +19,6 @@
 */
 
 #include "module.h"
-#include "commands.h"
 #include "misc.h"
 #include "special-vars.h"
 #include "settings.h"
@@ -31,7 +30,7 @@
 #include "nicklist.h"
 
 #include "bans.h"
-#include "irc.h"
+#include "irc-commands.h"
 #include "irc-servers.h"
 #include "irc-channels.h"
 #include "irc-queries.h"
@@ -43,28 +42,38 @@
    is required */
 #define LIST_MAX_CHANNELS_PASS 1000
 
+/* When /PARTing a channel, if there's more messages in output queue
+   than this, purge the output for channel. The idea behind this is that
+   if you accidentally pasted some large text and /PART the channel, the
+   text won't be fully pasted. Note that this counter is the whole size
+   of the output queue, not channel specific.. */
+#define MAX_COMMANDS_ON_PART_UNTIL_PURGE 10
+
 typedef struct {
 	IRC_CHANNEL_REC *channel;
 	char *ban;
-        int timeleft;
+	time_t unban_time;
 } KNOCKOUT_REC;
 
 static GString *tmpstr;
 static int knockout_tag;
 
 /* SYNTAX: NOTICE <targets> <message> */
-static void cmd_notice(const char *data, IRC_SERVER_REC *server)
+static void cmd_notice(const char *data, IRC_SERVER_REC *server,
+		       WI_ITEM_REC *item)
 {
 	char *target, *msg;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
-	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_GETREST, &target, &msg))
+	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_GETREST,
+			    &target, &msg))
 		return;
-	if (*target == '\0' || *msg == '\0') cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
+	if (strcmp(target, "*") == 0)
+		target = item == NULL ? NULL : item->name;
+	if (*target == '\0' || *msg == '\0')
+		cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
 
 	g_string_sprintf(tmpstr, "NOTICE %s :%s", target, msg);
 	irc_send_cmd_split(server, tmpstr->str, 2, server->max_msgs_in_cmd);
@@ -73,18 +82,21 @@ static void cmd_notice(const char *data, IRC_SERVER_REC *server)
 }
 
 /* SYNTAX: CTCP <targets> <ctcp command> [<ctcp data>] */
-static void cmd_ctcp(const char *data, IRC_SERVER_REC *server)
+static void cmd_ctcp(const char *data, IRC_SERVER_REC *server,
+		     WI_ITEM_REC *item)
 {
 	char *target, *ctcpcmd, *ctcpdata;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
-	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_GETREST, &target, &ctcpcmd, &ctcpdata))
+	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_GETREST,
+			    &target, &ctcpcmd, &ctcpdata))
 		return;
-	if (*target == '\0' || *ctcpcmd == '\0') cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
+	if (strcmp(target, "*") == 0)
+		target = item == NULL ? NULL : item->name;
+	if (*target == '\0' || *ctcpcmd == '\0')
+		cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
 
 	g_strup(ctcpcmd);
 	if (*ctcpdata == '\0')
@@ -97,18 +109,21 @@ static void cmd_ctcp(const char *data, IRC_SERVER_REC *server)
 }
 
 /* SYNTAX: NCTCP <targets> <ctcp command> [<ctcp data>] */
-static void cmd_nctcp(const char *data, IRC_SERVER_REC *server)
+static void cmd_nctcp(const char *data, IRC_SERVER_REC *server,
+		      WI_ITEM_REC *item)
 {
 	char *target, *ctcpcmd, *ctcpdata;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
-	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_GETREST, &target, &ctcpcmd, &ctcpdata))
+	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_GETREST,
+			    &target, &ctcpcmd, &ctcpdata))
 		return;
-	if (*target == '\0' || *ctcpcmd == '\0') cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
+	if (strcmp(target, "*") == 0)
+		target = item == NULL ? NULL : item->name;
+	if (*target == '\0' || *ctcpcmd == '\0')
+		cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
 
 	g_strup(ctcpcmd);
 	g_string_sprintf(tmpstr, "NOTICE %s :\001%s %s\001", target, ctcpcmd, ctcpdata);
@@ -118,20 +133,24 @@ static void cmd_nctcp(const char *data, IRC_SERVER_REC *server)
 }
 
 /* SYNTAX: PART [<channels>] [<message>] */
-static void cmd_part(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item)
+static void cmd_part(const char *data, IRC_SERVER_REC *server,
+		     WI_ITEM_REC *item)
 {
 	char *channame, *msg;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
-	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_OPTCHAN | PARAM_FLAG_GETREST, item, &channame, &msg))
+	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_GETREST |
+			    PARAM_FLAG_OPTCHAN, item, &channame, &msg))
 		return;
 	if (*channame == '\0') cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
 
 	if (*msg == '\0') msg = (char *) settings_get_str("part_message");
+
+        if (server->cmdcount > MAX_COMMANDS_ON_PART_UNTIL_PURGE)
+		irc_server_purge_output(server, channame);
+
 	irc_send_cmdv(server, *msg == '\0' ? "PART %s" : "PART %s :%s",
 		      channame, msg);
 
@@ -144,12 +163,11 @@ static void cmd_kick(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item
 	char *channame, *nicks, *reason;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
-	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_OPTCHAN | PARAM_FLAG_GETREST,
-			    item, &channame, &nicks, &reason))
+	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_GETREST |
+			    PARAM_FLAG_OPTCHAN, item,
+			    &channame, &nicks, &reason))
 		return;
 
 	if (*channame == '\0' || *nicks == '\0') cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
@@ -168,9 +186,7 @@ static void cmd_topic(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *ite
 	char *channame, *topic;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_OPTCHAN |
 			    PARAM_FLAG_OPTIONS | PARAM_FLAG_GETREST,
@@ -190,9 +206,7 @@ static void cmd_invite(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *it
 	char *nick, *channame;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 2, &nick, &channame))
 		return;
@@ -217,9 +231,7 @@ static void cmd_list(const char *data, IRC_SERVER_REC *server,
 	char *str;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 1 | PARAM_FLAG_OPTIONS |
 			    PARAM_FLAG_GETREST, "list", &optlist, &str))
@@ -232,9 +244,6 @@ static void cmd_list(const char *data, IRC_SERVER_REC *server,
 
 	irc_send_cmdv(server, "LIST %s", str);
 	cmd_params_free(free_arg);
-
-	/* add default redirection */
-	server_redirect_default((SERVER_REC *) server, "bogus command list");
 }
 
 /* SYNTAX: WHO [<nicks> | <channels> | **] */
@@ -244,9 +253,7 @@ static void cmd_who(const char *data, IRC_SERVER_REC *server,
 	char *channel, *rest;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_GETREST, &channel, &rest))
 		return;
@@ -265,9 +272,6 @@ static void cmd_who(const char *data, IRC_SERVER_REC *server,
 	irc_send_cmdv(server, *rest == '\0' ? "WHO %s" : "WHO %s %s",
 		      channel, rest);
 	cmd_params_free(free_arg);
-
-	/* add default redirection */
-	server_redirect_default((SERVER_REC *) server, "bogus command who");
 }
 
 static void cmd_names(const char *data, IRC_SERVER_REC *server,
@@ -277,9 +281,7 @@ static void cmd_names(const char *data, IRC_SERVER_REC *server,
 	char *channel;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 1 | PARAM_FLAG_OPTIONS |
 			    PARAM_FLAG_GETREST, "names", &optlist, &channel))
@@ -310,42 +312,18 @@ static void cmd_nick(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item
 
 	g_return_if_fail(data != NULL);
 
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 1, &nick))
 		return;
 
-	server->nick_changing = TRUE;
+	g_free(server->last_nick);
+	server->last_nick = g_strdup(nick);
+
 	irc_send_cmdv(server, "NICK %s", nick);
-
-	nick = g_strdup_printf("%s :%s", nick, nick);
-	server_redirect_event(SERVER(server), nick, 5,
-			      "event nick", "nickchange over", -1,
-			      "event 433", "nickchange over", 1,
-			      /* 437: ircnet = target unavailable,
-				      dalnet = banned in channel,
-				               can't change nick */
-			      "event 437", "nickchange over", -1,
-			      "event 432", "nickchange over", 1,
-			      "event 438", "nickchange over", 1, NULL);
-        g_free(nick);
 	cmd_params_free(free_arg);
 }
 
-static void sig_nickchange_over(IRC_SERVER_REC *server, const char *data,
-				const char *nick, const char *addr)
-{
-	char *signal;
-
-	server->nick_changing = FALSE;
-
-	signal = g_strconcat("event ", current_server_event, NULL);
-	g_strdown(signal+6);
-	signal_emit(signal, 4, server, data, nick, addr);
-	g_free(signal);
-}
-
 static char *get_redirect_nicklist(const char *nicks, int *free)
 {
 	char *str, *ret;
@@ -374,9 +352,7 @@ static void cmd_whois(const char *data, IRC_SERVER_REC *server,
 	void *free_arg;
 	int free_nick;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_OPTIONS,
 			    "whois", &optlist, &qserver, &query))
@@ -407,20 +383,22 @@ static void cmd_whois(const char *data, IRC_SERVER_REC *server,
 			event_402 = "whois event noserver";
 	}
 
-	server->whois_found = FALSE;
-	irc_send_cmd_split(server, tmpstr->str, 2, server->max_whois_in_cmd);
-
 	/* do automatic /WHOWAS if any of the nicks wasn't found */
 	query = get_redirect_nicklist(query, &free_nick);
 
         str = g_strconcat(qserver, " ", query, NULL);
-	server_redirect_event(SERVER(server), str, 2,
-			      "event 318", "event 318", 1,
-			      "event 402", event_402, 1,
-			      "event 401", "whois not found", 1,
-			      "event 311", "whois event", 1, NULL);
+	server_redirect_event(server, "whois", 1, str, TRUE,
+                              NULL,
+			      "event 318", "whois end",
+			      "event 402", event_402,
+			      "event 301", "whois away", /* 301 can come as a reply to /MSG, /WHOIS or /WHOWAS */
+			      "event 401", "whois not found",
+			      "event 311", "whois event", NULL);
         g_free(str);
 
+	server->whois_found = FALSE;
+	irc_send_cmd_split(server, tmpstr->str, 2, server->max_whois_in_cmd);
+
 	if (free_nick) g_free(query);
 	cmd_params_free(free_arg);
 }
@@ -439,17 +417,26 @@ static void sig_whois_not_found(IRC_SERVER_REC *server, const char *data)
 	g_return_if_fail(data != NULL);
 
 	params = event_get_params(data, 2, NULL, &nick);
-	irc_send_cmdv(server, "WHOWAS %s 1", nick);
 
 	server->whowas_found = FALSE;
-	server_redirect_event(SERVER(server), nick, 1,
-			      "event 369", "whowas event end", 1,
-			      "event 314", "whowas event", 1,
-			      "event 406", "event empty", 1, NULL);
+	server_redirect_event(server, "whowas", 1, nick, -1, NULL,
+			      "event 314", "whowas event",
+			      "event 369", "whowas event end",
+			      "event 406", "event empty", NULL);
+	irc_send_cmdv(server, "WHOWAS %s 1", nick);
+
 	g_free(params);
 }
 
-static void event_whowas(IRC_SERVER_REC *server, const char *data, const char *nick, const char *addr)
+static void event_end_of_whois(IRC_SERVER_REC *server, const char *data,
+			       const char *nick, const char *addr)
+{
+	signal_emit("event 318", 4, server, data, nick, addr);
+	server->whois_found = FALSE;
+}
+
+static void event_whowas(IRC_SERVER_REC *server, const char *data,
+			 const char *nick, const char *addr)
 {
 	server->whowas_found = TRUE;
 	signal_emit("event 314", 4, server, data, nick, addr);
@@ -458,27 +445,26 @@ static void event_whowas(IRC_SERVER_REC *server, const char *data, const char *n
 /* SYNTAX: WHOWAS [<nicks> [<count>]] */
 static void cmd_whowas(const char *data, IRC_SERVER_REC *server)
 {
-	char *nicks, *count;
+	char *nicks, *count, *nicks_redir;
 	void *free_arg;
 	int free_nick;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 2, &nicks, &count))
 		return;
 	if (*nicks == '\0') nicks = server->nick;
 
+	nicks_redir = get_redirect_nicklist(nicks, &free_nick);
+	server_redirect_event(server, "whowas", 1, nicks_redir, -1, NULL,
+			      "event 301", "whowas away", /* 301 can come as a reply to /MSG, /WHOIS or /WHOWAS */
+			      "event 314", "whowas event", NULL);
+	if (free_nick) g_free(nicks_redir);
+
 	server->whowas_found = FALSE;
 	irc_send_cmdv(server, *count == '\0' ? "WHOWAS %s" :
 		      "WHOWAS %s %s", nicks, count);
 
-	nicks = get_redirect_nicklist(nicks, &free_nick);
-	server_redirect_event(SERVER(server), nicks, 1,
-			      "event 369", "event 369", 1,
-			      "event 314", "whowas event", 1, NULL);
-	if (free_nick) g_free(nicks);
 	cmd_params_free(free_arg);
 }
 
@@ -488,9 +474,7 @@ static void cmd_ping(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item
 	GTimeVal tv;
         char *str;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (*data == '\0' || strcmp(data, "*") == 0) {
 		if (!IS_IRC_ITEM(item))
@@ -511,10 +495,13 @@ static void server_send_away(IRC_SERVER_REC *server, const char *reason)
 	if (!IS_IRC_SERVER(server))
 		return;
 
-	g_free_not_null(server->away_reason);
-	server->away_reason = g_strdup(reason);
+	if (*reason != '\0' || server->usermode_away) {
+		g_free_and_null(server->away_reason);
+                if (*reason != '\0')
+			server->away_reason = g_strdup(reason);
 
-	irc_send_cmdv(server, "AWAY :%s", reason);
+		irc_send_cmdv(server, "AWAY :%s", reason);
+	}
 }
 
 /* SYNTAX: AWAY [-one | -all] [<reason>] */
@@ -524,9 +511,7 @@ static void cmd_away(const char *data, IRC_SERVER_REC *server)
 	char *reason;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 1 | PARAM_FLAG_OPTIONS |
 			    PARAM_FLAG_GETREST, "away", &optlist, &reason)) return;
@@ -539,23 +524,10 @@ static void cmd_away(const char *data, IRC_SERVER_REC *server)
 	cmd_params_free(free_arg);
 }
 
-/* SYNTAX: DEOP <nicks> */
-static void cmd_deop(const char *data, IRC_SERVER_REC *server)
-{
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
-
-	if (*data == '\0')
-		irc_send_cmdv(server, "MODE %s -o", server->nick);
-}
-
 /* SYNTAX: SCONNECT <new server> [[<port>] <existing server>] */
 static void cmd_sconnect(const char *data, IRC_SERVER_REC *server)
 {
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 	if (*data == '\0') cmd_return_error(CMDERR_NOT_ENOUGH_PARAMS);
 
 	irc_send_cmdv(server, "CONNECT %s", data);
@@ -564,13 +536,25 @@ static void cmd_sconnect(const char *data, IRC_SERVER_REC *server)
 /* SYNTAX: QUOTE <data> */
 static void cmd_quote(const char *data, IRC_SERVER_REC *server)
 {
-	g_return_if_fail(data != NULL);
-	if (server == NULL || !IS_IRC_SERVER(server))
+	if (server != NULL && !IS_IRC_SERVER(server))
+		return;
+	if (server == NULL || server->connect_time == 0)
 		cmd_return_error(CMDERR_NOT_CONNECTED);
 
 	irc_send_cmd(server, data);
 }
 
+/* SYNTAX: RAWQUOTE <data> */
+static void cmd_rawquote(const char *data, IRC_SERVER_REC *server)
+{
+	if (server != NULL && !IS_IRC_SERVER(server))
+		return;
+	if (server == NULL || server->connect_time == 0)
+		cmd_return_error(CMDERR_NOT_CONNECTED);
+
+	irc_send_cmd_full(server, data, FALSE, FALSE, TRUE);
+}
+
 static void cmd_wall_hash(gpointer key, NICK_REC *nick, GSList **nicks)
 {
 	if (nick->op) *nicks = g_slist_append(*nicks, nick);
@@ -584,9 +568,7 @@ static void cmd_wait(const char *data, IRC_SERVER_REC *server)
 	void *free_arg;
 	int n;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 1 | PARAM_FLAG_OPTIONS |
 			    PARAM_FLAG_UNKNOWN_OPTIONS | PARAM_FLAG_GETREST,
@@ -621,9 +603,7 @@ static void cmd_wall(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item
 	IRC_CHANNEL_REC *chanrec;
 	GSList *tmp, *nicks;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_OPTCHAN |
 			    PARAM_FLAG_GETREST, item, &channame, &msg))
@@ -654,28 +634,20 @@ static void cmd_wall(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item
 	cmd_params_free(free_arg);
 }
 
-/* SYNTAX: CYCLE [<channel>] [<message>] */
-static void cmd_cycle(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item)
+/* SYNTAX: WALLCHOPS <channel> <message> */
+static void cmd_wallchops(const char *data, IRC_SERVER_REC *server, WI_ITEM_REC *item)
 {
-	IRC_CHANNEL_REC *chanrec;
 	char *channame, *msg;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
-	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_OPTCHAN, item, &channame, &msg))
+	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_OPTCHAN |
+			    PARAM_FLAG_GETREST, item, &channame, &msg))
 		return;
-	if (*channame == '\0') cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
+	if (*msg == '\0') cmd_param_error(CMDERR_NOT_ENOUGH_PARAMS);
 
-	chanrec = irc_channel_find(server, channame);
-	if (chanrec == NULL) cmd_param_error(CMDERR_CHAN_NOT_FOUND);
-
-	irc_send_cmdv(server, *msg == '\0' ? "PART %s" : "PART %s :%s",
-		      channame, msg);
-	irc_send_cmdv(server, chanrec->key == NULL ? "JOIN %s" : "JOIN %s %s",
-		      channame, chanrec->key);
+	irc_send_cmdv(server, "WALLCHOPS %s :%s", channame, msg);
 
 	cmd_params_free(free_arg);
 }
@@ -685,13 +657,11 @@ static void cmd_kickban(const char *data, IRC_SERVER_REC *server,
 			WI_ITEM_REC *item)
 {
         IRC_CHANNEL_REC *chanrec;
-	char *channel, *nicks, *reason, *str;
+	char *channel, *nicks, *reason, *kickcmd, *bancmd;
         char **nicklist, *spacenicks;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 3 | PARAM_FLAG_OPTCHAN | PARAM_FLAG_GETREST,
 			    item, &channel, &nicks, &reason))
@@ -710,15 +680,20 @@ static void cmd_kickban(const char *data, IRC_SERVER_REC *server,
         spacenicks = g_strjoinv(" ", nicklist);
 	g_strfreev(nicklist);
 
-	str = g_strdup_printf("%s %s", chanrec->name, spacenicks);
-	signal_emit("command ban", 3, str, server, channel);
-	g_free(str);
-
-	str = g_strdup_printf("%s %s %s", chanrec->name, nicks, reason);
-	signal_emit("command kick", 3, str, server, channel);
-	g_free(str);
-
+	kickcmd = g_strdup_printf("%s %s %s", chanrec->name, nicks, reason);
+	bancmd = g_strdup_printf("%s %s", chanrec->name, spacenicks);
         g_free(spacenicks);
+
+        if (settings_get_bool("kick_first_on_kickban")) {
+		signal_emit("command kick", 3, kickcmd, server, chanrec);
+		signal_emit("command ban", 3, bancmd, server, chanrec);
+	} else {
+		signal_emit("command ban", 3, bancmd, server, chanrec);
+		signal_emit("command kick", 3, kickcmd, server, chanrec);
+	}
+	g_free(kickcmd);
+	g_free(bancmd);
+
 	cmd_params_free(free_arg);
 }
 
@@ -733,24 +708,19 @@ static void knockout_destroy(IRC_SERVER_REC *server, KNOCKOUT_REC *rec)
 static void knockout_timeout_server(IRC_SERVER_REC *server)
 {
 	GSList *tmp, *next;
-	time_t t;
+	time_t now;
 
 	g_return_if_fail(server != NULL);
 
 	if (!IS_IRC_SERVER(server))
 		return;
 
-	t = server->knockout_lastcheck == 0 ? 0 :
-		time(NULL)-server->knockout_lastcheck;
-	server->knockout_lastcheck = time(NULL);
-
+        now = time(NULL);
 	for (tmp = server->knockoutlist; tmp != NULL; tmp = next) {
 		KNOCKOUT_REC *rec = tmp->data;
 
 		next = tmp->next;
-		if (rec->timeleft > t)
-			rec->timeleft -= t;
-		else {
+		if (rec->unban_time <= now) {
 			/* timeout, unban. */
 			ban_remove(rec->channel, rec->ban);
 			knockout_destroy(server, rec);
@@ -769,12 +739,13 @@ static void cmd_knockout(const char *data, IRC_SERVER_REC *server,
 			 IRC_CHANNEL_REC *channel)
 {
 	KNOCKOUT_REC *rec;
-	char *nicks, *reason, *timeoutstr, *str;
+	char *nicks, *reason, *timeoutstr, *kickcmd, *bancmd;
         char **nicklist, *spacenicks, *banmasks;
 	void *free_arg;
 	int timeleft;
 
-	g_return_if_fail(data != NULL);
+        CMD_IRC_SERVER(server);
+
 	if (!IS_IRC_CHANNEL(channel))
 		cmd_return_error(CMDERR_NOT_JOINED);
 	if (!channel->wholist)
@@ -803,22 +774,28 @@ static void cmd_knockout(const char *data, IRC_SERVER_REC *server,
 	banmasks = ban_get_masks(channel, spacenicks, 0);
 	g_free(spacenicks);
 
-	if (*banmasks != '\0') {
-		str = g_strdup_printf("%s %s", channel->name, banmasks);
-		signal_emit("command ban", 3, str, server, channel);
-		g_free(str);
+	kickcmd = g_strdup_printf("%s %s %s", channel->name, nicks, reason);
+	bancmd = *banmasks == '\0'? NULL :
+		g_strdup_printf("%s %s", channel->name, banmasks);
+	
+        if (settings_get_bool("kick_first_on_kickban")) {
+		signal_emit("command kick", 3, kickcmd, server, channel);
+		if (bancmd != NULL)
+			signal_emit("command ban", 3, bancmd, server, channel);
+	} else {
+		if (bancmd != NULL)
+			signal_emit("command ban", 3, bancmd, server, channel);
+		signal_emit("command kick", 3, kickcmd, server, channel);
 	}
-
-	str = g_strdup_printf("%s %s %s", channel->name, nicks, reason);
-	signal_emit("command kick", 3, str, server, channel);
-	g_free(str);
+	g_free(kickcmd);
+	g_free_not_null(bancmd);
 
 	if (*banmasks == '\0')
 		g_free(banmasks);
 	else {
 		/* create knockout record */
 		rec = g_new(KNOCKOUT_REC, 1);
-		rec->timeleft = timeleft;
+		rec->unban_time = time(NULL)+timeleft;
 		rec->channel = channel;
 		rec->ban = banmasks;
 
@@ -828,6 +805,22 @@ static void cmd_knockout(const char *data, IRC_SERVER_REC *server,
 	cmd_params_free(free_arg);
 }
 
+/* SYNTAX: SERVER PURGE [<target>] */
+static void cmd_server_purge(const char *data, IRC_SERVER_REC *server)
+{
+        char *target;
+	void *free_arg;
+
+        CMD_IRC_SERVER(server);
+
+	if (!cmd_get_params(data, &free_arg, 1, &target))
+		return;
+
+	irc_server_purge_output(server, *target == '\0' ? NULL : target);
+
+	cmd_params_free(free_arg);
+}
+
 /* destroy all knockouts in server */
 static void sig_server_disconnected(IRC_SERVER_REC *server)
 {
@@ -863,9 +856,7 @@ static void cmd_oper(const char *data, IRC_SERVER_REC *server)
 	char *nick, *password;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
         /* asking for password is handled by fe-common */
 	if (!cmd_get_params(data, &free_arg, 2, &nick, &password))
@@ -879,9 +870,7 @@ static void cmd_oper(const char *data, IRC_SERVER_REC *server)
 /* SYNTAX: UNSILENCE <nick!user@host> */
 static void cmd_unsilence(const char *data, IRC_SERVER_REC *server)
 {
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (*data == '\0') 
 		cmd_return_error(CMDERR_NOT_ENOUGH_PARAMS);
@@ -891,9 +880,7 @@ static void cmd_unsilence(const char *data, IRC_SERVER_REC *server)
 
 static void command_self(const char *data, IRC_SERVER_REC *server)
 {
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	irc_send_cmdv(server, *data == '\0' ? "%s" : "%s %s", current_command, data);
 }
@@ -913,9 +900,7 @@ static void command_2self(const char *data, IRC_SERVER_REC *server)
 	char *target, *text;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (!IS_IRC_SERVER(server) || !server->connected)
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_GETREST, &target, &text))
 		return;
@@ -924,48 +909,6 @@ static void command_2self(const char *data, IRC_SERVER_REC *server)
 	cmd_params_free(free_arg);
 }
 
-static void sig_connected(IRC_SERVER_REC *server)
-{
-	g_return_if_fail(server != NULL);
-
-	/* FIXME: these two aren't probably needed? this whole redirection
-	   thing might need some rethinking :) */
-	/* WHOIS */
-	/*server_redirect_init(SERVER(server), "", 2,
-			     "event 318", "event 402", "event 401",
-			     "event 301", "event 311", "event 312", "event 313",
-			     "event 317", "event 319", NULL);*/
-
-	/* NICK */
-	/*server_redirect_init(SERVER(server), "", 5,
-			     "event nick", "event 433", "event 437",
-			     "event 432", "event 438", NULL);*/
-
-	/* problem (doesn't really apply currently since there's no GUI):
-
-	   second argument of server_redirect_init() is the command that
-	   generates the redirection automatically when it's called, but the
-	   command handler doesn't really know about the redirection itself.
-
-	   every time the command is called, this redirection is generated.
-	   this is a problem if the redirection is wanted sometimes but not
-	   always. for example /WHO #channel could create a window with a
-	   list of people in channel redirecting WHO's events to it's own use,
-	   but /WHO -nogui #channel would use the default WHO handler which
-	   doesn't know anything about redirection. with GUI /WHO the
-	   redirection would be done twice then..
-
-	   so the kludgy workaround currently is this: make the default
-	   handler handle the redirection always.. when default WHO/LIST
-	   handler is called, they call
-	   server_redirect_default("bogus command who") or ..list..
-
-	   this is really a problem if some script/plugin wants to override
-           some default command to use redirections.. */
-	server_redirect_init(SERVER(server), "bogus command who", 2, "event 401", "event 315", "event 352", NULL);
-	server_redirect_init(SERVER(server), "bogus command list", 1, "event 321", "event 322", "event 323", NULL);
-}
-
 void irc_commands_init(void)
 {
 	tmpstr = g_string_new(NULL);
@@ -973,95 +916,94 @@ void irc_commands_init(void)
 	settings_add_str("misc", "part_message", "");
 	settings_add_int("misc", "knockout_time", 300);
 	settings_add_str("misc", "wall_format", "[Wall/$0] $1-");
+	settings_add_bool("misc", "kick_first_on_kickban", FALSE);
 
 	knockout_tag = g_timeout_add(KNOCKOUT_TIMECHECK, (GSourceFunc) knockout_timeout, NULL);
 
-	command_bind("notice", NULL, (SIGNAL_FUNC) cmd_notice);
-	command_bind("ctcp", NULL, (SIGNAL_FUNC) cmd_ctcp);
-	command_bind("nctcp", NULL, (SIGNAL_FUNC) cmd_nctcp);
-	command_bind("part", NULL, (SIGNAL_FUNC) cmd_part);
-	command_bind("kick", NULL, (SIGNAL_FUNC) cmd_kick);
-	command_bind("topic", NULL, (SIGNAL_FUNC) cmd_topic);
-	command_bind("invite", NULL, (SIGNAL_FUNC) cmd_invite);
-	command_bind("list", NULL, (SIGNAL_FUNC) cmd_list);
-	command_bind("who", NULL, (SIGNAL_FUNC) cmd_who);
-	command_bind("names", NULL, (SIGNAL_FUNC) cmd_names);
-	command_bind("nick", NULL, (SIGNAL_FUNC) cmd_nick);
+	command_bind_irc("notice", NULL, (SIGNAL_FUNC) cmd_notice);
+	command_bind_irc("ctcp", NULL, (SIGNAL_FUNC) cmd_ctcp);
+	command_bind_irc("nctcp", NULL, (SIGNAL_FUNC) cmd_nctcp);
+	command_bind_irc("part", NULL, (SIGNAL_FUNC) cmd_part);
+	command_bind_irc("kick", NULL, (SIGNAL_FUNC) cmd_kick);
+	command_bind_irc("topic", NULL, (SIGNAL_FUNC) cmd_topic);
+	command_bind_irc("invite", NULL, (SIGNAL_FUNC) cmd_invite);
+	command_bind_irc("list", NULL, (SIGNAL_FUNC) cmd_list);
+	command_bind_irc("who", NULL, (SIGNAL_FUNC) cmd_who);
+	command_bind_irc("names", NULL, (SIGNAL_FUNC) cmd_names);
+	command_bind_irc("nick", NULL, (SIGNAL_FUNC) cmd_nick);
 	/* SYNTAX: NOTE <command> [&<password>] [+|-<flags>] [<arguments>] */
-	command_bind("note", NULL, (SIGNAL_FUNC) command_self);
-	command_bind("whois", NULL, (SIGNAL_FUNC) cmd_whois);
-	command_bind("whowas", NULL, (SIGNAL_FUNC) cmd_whowas);
-	command_bind("ping", NULL, (SIGNAL_FUNC) cmd_ping);
+	command_bind_irc("note", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("whois", NULL, (SIGNAL_FUNC) cmd_whois);
+	command_bind_irc("whowas", NULL, (SIGNAL_FUNC) cmd_whowas);
+	command_bind_irc("ping", NULL, (SIGNAL_FUNC) cmd_ping);
 	/* SYNTAX: KILL <nick> <reason> */
-	command_bind("kill", NULL, (SIGNAL_FUNC) command_2self);
-	command_bind("away", NULL, (SIGNAL_FUNC) cmd_away);
+	command_bind_irc("kill", NULL, (SIGNAL_FUNC) command_2self);
+	command_bind_irc("away", NULL, (SIGNAL_FUNC) cmd_away);
 	/* SYNTAX: ISON <nicks> */
-	command_bind("ison", NULL, (SIGNAL_FUNC) command_1self);
+	command_bind_irc("ison", NULL, (SIGNAL_FUNC) command_1self);
 	/* SYNTAX: ADMIN [<server>|<nickname>] */
-	command_bind("admin", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("admin", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: INFO [<server>] */
-	command_bind("info", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("info", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: LINKS [[<server>] <mask>] */
-	command_bind("links", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("links", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: LUSERS [<server mask> [<remote server>]] */
-	command_bind("lusers", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("lusers", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: MAP */
-	command_bind("map", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("map", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: MOTD [<server>|<nick>] */
-	command_bind("motd", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("motd", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: REHASH */
-	command_bind("rehash", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("rehash", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: STATS <type> [<server>] */
-	command_bind("stats", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("stats", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: TIME [<server>|<nick>] */
-	command_bind("time", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("time", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: TRACE [<server>|<nick>] */
-	command_bind("trace", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("trace", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: VERSION [<server>|<nick>] */
-	command_bind("version", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("version", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: SERVLIST [<server mask>] */
-	command_bind("servlist", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("servlist", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: SILENCE [[+|-]<nick!user@host>]
 	           SILENCE [<nick>] */
-	command_bind("silence", NULL, (SIGNAL_FUNC) command_self);
-	command_bind("unsilence", NULL, (SIGNAL_FUNC) cmd_unsilence);
-	command_bind("sconnect", NULL, (SIGNAL_FUNC) cmd_sconnect);
+	command_bind_irc("silence", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("unsilence", NULL, (SIGNAL_FUNC) cmd_unsilence);
+	command_bind_irc("sconnect", NULL, (SIGNAL_FUNC) cmd_sconnect);
 	/* SYNTAX: SQUERY <service> [<commands>] */
-	command_bind("squery", NULL, (SIGNAL_FUNC) command_2self);
-	command_bind("deop", NULL, (SIGNAL_FUNC) cmd_deop);
+	command_bind_irc("squery", NULL, (SIGNAL_FUNC) command_2self);
 	/* SYNTAX: DIE */
-	command_bind("die", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("die", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: HASH */
-	command_bind("hash", NULL, (SIGNAL_FUNC) command_self);
-	command_bind("oper", NULL, (SIGNAL_FUNC) cmd_oper);
+	command_bind_irc("hash", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("oper", NULL, (SIGNAL_FUNC) cmd_oper);
 	/* SYNTAX: RESTART */
-	command_bind("restart", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("restart", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: RPING <server> */
-	command_bind("rping", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("rping", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: SQUIT <server>|<mask> <reason> */
-	command_bind("squit", NULL, (SIGNAL_FUNC) command_2self);
+	command_bind_irc("squit", NULL, (SIGNAL_FUNC) command_2self);
 	/* SYNTAX: UPING <server> */
-	command_bind("uping", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("uping", NULL, (SIGNAL_FUNC) command_self);
 	/* SYNTAX: USERHOST <nicks> */
-	command_bind("userhost", NULL, (SIGNAL_FUNC) command_self);
-	command_bind("quote", NULL, (SIGNAL_FUNC) cmd_quote);
-	command_bind("wall", NULL, (SIGNAL_FUNC) cmd_wall);
-	command_bind("wait", NULL, (SIGNAL_FUNC) cmd_wait);
+	command_bind_irc("userhost", NULL, (SIGNAL_FUNC) command_self);
+	command_bind_irc("quote", NULL, (SIGNAL_FUNC) cmd_quote);
+	command_bind_irc("rawquote", NULL, (SIGNAL_FUNC) cmd_rawquote);
+	command_bind_irc("wall", NULL, (SIGNAL_FUNC) cmd_wall);
+	command_bind_irc("wallchops", NULL, (SIGNAL_FUNC) cmd_wallchops);
+	command_bind_irc("wait", NULL, (SIGNAL_FUNC) cmd_wait);
 	/* SYNTAX: WALLOPS <message> */
-	command_bind("wallops", NULL, (SIGNAL_FUNC) command_1self);
-	/* SYNTAX: WALLCHOPS <channel> <message> */
-	command_bind("wallchops", NULL, (SIGNAL_FUNC) command_2self);
-	command_bind("cycle", NULL, (SIGNAL_FUNC) cmd_cycle);
-	command_bind("kickban", NULL, (SIGNAL_FUNC) cmd_kickban);
-	command_bind("knockout", NULL, (SIGNAL_FUNC) cmd_knockout);
+	command_bind_irc("wallops", NULL, (SIGNAL_FUNC) command_1self);
+	command_bind_irc("kickban", NULL, (SIGNAL_FUNC) cmd_kickban);
+	command_bind_irc("knockout", NULL, (SIGNAL_FUNC) cmd_knockout);
+	command_bind_irc("server purge", NULL, (SIGNAL_FUNC) cmd_server_purge);
 
 	signal_add("channel destroyed", (SIGNAL_FUNC) sig_channel_destroyed);
 	signal_add("server disconnected", (SIGNAL_FUNC) sig_server_disconnected);
-	signal_add("nickchange over", (SIGNAL_FUNC) sig_nickchange_over);
 	signal_add("whois not found", (SIGNAL_FUNC) sig_whois_not_found);
 	signal_add("whois event", (SIGNAL_FUNC) event_whois);
+	signal_add("whois end", (SIGNAL_FUNC) event_end_of_whois);
 	signal_add("whowas event", (SIGNAL_FUNC) event_whowas);
-	signal_add("server connected", (SIGNAL_FUNC) sig_connected);
 
 	command_set_options("connect", "+ircnet");
 	command_set_options("topic", "delete");
@@ -1108,7 +1050,6 @@ void irc_commands_deinit(void)
 	command_unbind("unsilence", (SIGNAL_FUNC) cmd_unsilence);
 	command_unbind("sconnect", (SIGNAL_FUNC) cmd_sconnect);
 	command_unbind("squery", (SIGNAL_FUNC) command_2self);
-	command_unbind("deop", (SIGNAL_FUNC) cmd_deop);
 	command_unbind("die", (SIGNAL_FUNC) command_self);
 	command_unbind("hash", (SIGNAL_FUNC) command_self);
 	command_unbind("oper", (SIGNAL_FUNC) cmd_oper);
@@ -1118,20 +1059,21 @@ void irc_commands_deinit(void)
 	command_unbind("uping", (SIGNAL_FUNC) command_self);
 	command_unbind("userhost", (SIGNAL_FUNC) command_self);
 	command_unbind("quote", (SIGNAL_FUNC) cmd_quote);
+	command_unbind("rawquote", (SIGNAL_FUNC) cmd_rawquote);
 	command_unbind("wall", (SIGNAL_FUNC) cmd_wall);
+	command_unbind("wallchops", (SIGNAL_FUNC) cmd_wallchops);
 	command_unbind("wait", (SIGNAL_FUNC) cmd_wait);
 	command_unbind("wallops", (SIGNAL_FUNC) command_1self);
-	command_unbind("wallchops", (SIGNAL_FUNC) command_2self);
-	command_unbind("cycle", (SIGNAL_FUNC) cmd_cycle);
 	command_unbind("kickban", (SIGNAL_FUNC) cmd_kickban);
 	command_unbind("knockout", (SIGNAL_FUNC) cmd_knockout);
+	command_unbind("server purge", (SIGNAL_FUNC) cmd_server_purge);
+
 	signal_remove("channel destroyed", (SIGNAL_FUNC) sig_channel_destroyed);
 	signal_remove("server disconnected", (SIGNAL_FUNC) sig_server_disconnected);
-	signal_remove("nickchange over", (SIGNAL_FUNC) sig_nickchange_over);
 	signal_remove("whois not found", (SIGNAL_FUNC) sig_whois_not_found);
 	signal_remove("whois event", (SIGNAL_FUNC) event_whois);
+	signal_remove("whois end", (SIGNAL_FUNC) event_end_of_whois);
 	signal_remove("whowas event", (SIGNAL_FUNC) event_whowas);
-	signal_remove("server connected", (SIGNAL_FUNC) sig_connected);
 
 	g_string_free(tmpstr, TRUE);
 }
