@@ -20,10 +20,12 @@
 
 #include "module.h"
 #include "signals.h"
-#include "commands.h"
 #include "settings.h"
 
-#include "irc.h"
+#include "irc-commands.h"
+#include "irc-servers.h"
+#include "irc-channels.h"
+#include "servers-redirect.h"
 #include "modes.h"
 #include "mode-lists.h"
 #include "nicklist.h"
@@ -211,14 +213,9 @@ void parse_channel_modes(IRC_CHANNEL_REC *channel, const char *setby,
 	curmode = cmd_get_param(&modestr);
 	while (*curmode != '\0') {
 		if (HAS_MODE_ARG(type, *curmode)) {
-			/* get the argument for the mode. since we're
-			   expecting argument, ignore the mode if there's
-			   no argument (shouldn't happen). */
+			/* get the argument for the mode. NOTE: We don't
+			   get the +k's argument when joining to channel. */
 			arg = cmd_get_param(&modestr);
-			if (*arg == '\0') {
-				curmode++;
-				continue;
-			}
 		} else {
 			arg = NULL;
 		}
@@ -235,21 +232,8 @@ void parse_channel_modes(IRC_CHANNEL_REC *channel, const char *setby,
 			else
 				banlist_remove(channel, arg);
 			break;
-		case 'e':
-			if (type == '+')
-				banlist_exception_add(channel, arg, setby,
-						      time(NULL));
-			else
-				banlist_exception_remove(channel, arg);
-			break;
-		case 'I':
-			if (type == '+')
-				invitelist_add(channel, arg);
-			else
-				invitelist_remove(channel, arg);
-			break;
-
 		case 'o':
+		case 'O': /* channel owner in !channels */
 			if (g_strcasecmp(channel->server->nick, arg) == 0)
 				channel->chanop = type == '+';
 			nick_mode_change(channel, arg, '@', type);
@@ -266,10 +250,22 @@ void parse_channel_modes(IRC_CHANNEL_REC *channel, const char *setby,
 			channel->limit = type == '-' ? 0 : atoi(arg);
 			break;
 		case 'k':
+			if (*arg == '\0' && type == '+') {
+				arg = channel->key != NULL ? channel->key :
+					"???";
+			}
 			mode_set_arg(newmode, type, 'k', arg);
-			g_free_and_null(channel->key);
-			if (type == '+')
-				channel->key = g_strdup(arg);
+
+			if (arg != channel->key) {
+				g_free_and_null(channel->key);
+				if (type == '+')
+					channel->key = g_strdup(arg);
+			}
+			break;
+		case 'e':
+		case 'I':
+		case 'q':
+			/* Don't set it as channel mode */
 			break;
 
 		default:
@@ -417,17 +413,47 @@ static void event_unaway(IRC_SERVER_REC *server, const char *data)
 	signal_emit("away mode changed", 1, server);
 }
 
-void channel_set_singlemode(IRC_SERVER_REC *server, const char *channel,
-			    const char *nicks, const char *mode)
+static void sig_req_usermode_change(IRC_SERVER_REC *server, const char *data,
+				    const char *nick, const char *addr)
+{
+	char *params, *target, *mode;
+
+	g_return_if_fail(data != NULL);
+
+	params = event_get_params(data, 2 | PARAM_FLAG_GETREST,
+				  &target, &mode);
+	if (!ischannel(*target)) {
+                /* we requested a user mode change, save this */
+		mode = modes_join(server->wanted_usermode, mode, FALSE);
+                g_free_not_null(server->wanted_usermode);
+		server->wanted_usermode = mode;
+	}
+
+	g_free(params);
+
+	signal_emit("event mode", 4, server, data, nick, addr);
+}
+
+void channel_set_singlemode(IRC_CHANNEL_REC *channel, const char *nicks,
+			    const char *mode)
 {
 	GString *str;
 	int num, modepos;
 	char **nick, **nicklist;
 
-	g_return_if_fail(IS_IRC_SERVER(server));
-	g_return_if_fail(channel != NULL && nicks != NULL && mode != NULL);
+	g_return_if_fail(IS_IRC_CHANNEL(channel));
+	g_return_if_fail(nicks != NULL && mode != NULL);
 	if (*nicks == '\0') return;
 
+	if (!channel->chanop) {
+                /* not op - can we do anything? */
+		if (channel->ownnick == NULL || !channel->ownnick->halfop)
+			return; /* not even halfop, abort */
+
+		if (mode[0] != '\0' && mode[1] == 'o')
+                        return; /* halfops can't op/deop */
+	}
+
 	num = modepos = 0;
 	str = g_string_new(NULL);
 
@@ -438,7 +464,8 @@ void channel_set_singlemode(IRC_SERVER_REC *server, const char *channel,
 
 		if (num == 0)
 		{
-			g_string_sprintf(str, "MODE %s %s", channel, mode);
+			g_string_sprintf(str, "MODE %s %s",
+					 channel->name, mode);
 			modepos = str->len;
 		} else {
 			/* insert the mode string */
@@ -447,13 +474,13 @@ void channel_set_singlemode(IRC_SERVER_REC *server, const char *channel,
 
 		g_string_sprintfa(str, " %s", *nick);
 
-		if (++num == server->max_modes_in_cmd) {
+		if (++num == channel->server->max_modes_in_cmd) {
 			/* max. modes / command reached, send to server */
-			irc_send_cmd(server, str->str);
+			irc_send_cmd(channel->server, str->str);
 			num = 0;
 		}
 	}
-	if (num > 0) irc_send_cmd(server, str->str);
+	if (num > 0) irc_send_cmd(channel->server, str->str);
 
 	g_strfreev(nicklist);
 	g_string_free(str, TRUE);
@@ -504,7 +531,18 @@ void channel_set_mode(IRC_SERVER_REC *server, const char *channel,
 
 			count++;
 			arg = cmd_get_param(&modestr);
-			if (*arg != '\0') g_string_sprintfa(targs, " %s", arg);
+			if (*arg == '\0' && type == '-' && *curmode == 'k') {
+				/* "/mode #channel -k" - no reason why it
+				   shouldn't work really, so append the key */
+				IRC_CHANNEL_REC *chanrec;
+
+				chanrec = irc_channel_find(server, channel);
+				if (chanrec != NULL && chanrec->key != NULL)
+                                        arg = chanrec->key;
+			}
+
+			if (*arg != '\0')
+				g_string_sprintfa(targs, " %s", arg);
 		}
 	}
 
@@ -518,15 +556,17 @@ void channel_set_mode(IRC_SERVER_REC *server, const char *channel,
 	g_free(orig);
 }
 
-static void get_wildcard_nicks(GString *output, const char *mask,
-			       IRC_CHANNEL_REC *channel, int op, int voice)
+static int get_wildcard_nicks(GString *output, const char *mask,
+			      IRC_CHANNEL_REC *channel, int op, int voice)
 {
 	GSList *nicks, *tmp;
+        int count;
 
-	g_return_if_fail(output != NULL);
-	g_return_if_fail(mask != NULL);
-	g_return_if_fail(IS_IRC_CHANNEL(channel));
+	g_return_val_if_fail(output != NULL, 0);
+	g_return_val_if_fail(mask != NULL, 0);
+	g_return_val_if_fail(IS_IRC_CHANNEL(channel), 0);
 
+        count = 0;
 	nicks = nicklist_find_multiple(CHANNEL(channel), mask);
 	for (tmp = nicks; tmp != NULL; tmp = tmp->next) {
 		NICK_REC *rec = tmp->data;
@@ -539,95 +579,130 @@ static void get_wildcard_nicks(GString *output, const char *mask,
 			continue;
 
 		g_string_sprintfa(output, "%s ", rec->nick);
+                count++;
 	}
 	g_slist_free(nicks);
+
+        return count;
 }
 
-static char *get_nicks(IRC_CHANNEL_REC *channel,
-		       const char *data, int op, int voice)
+static char *get_nicks(IRC_SERVER_REC *server, WI_ITEM_REC *item,
+		       const char *data, int op, int voice,
+		       IRC_CHANNEL_REC **ret_channel)
 {
+        IRC_CHANNEL_REC *channel;
         GString *str;
-	char **matches, **match, *ret;
+        GHashTable *optlist;
+	char **matches, **match, *ret, *channame, *nicks;
+	void *free_arg;
+        int count, max_modes;
 
-	g_return_val_if_fail(channel != NULL, NULL);
-	g_return_val_if_fail(data != NULL, NULL);
-	if (*data == '\0') return NULL;
+	if (!cmd_get_params(data, &free_arg, 2 | PARAM_FLAG_GETREST |
+			    PARAM_FLAG_OPTIONS | PARAM_FLAG_OPTCHAN_NAME,
+			    item, "op", &optlist, &channame, &nicks))
+		return NULL;
+
+	if (*nicks == '\0')
+		return NULL;
+
+	channel = irc_channel_find(server, channame);
+	if (channel == NULL) {
+		cmd_params_free(free_arg);
+		return NULL;
+	}
 
 	str = g_string_new(NULL);
-	matches = g_strsplit(data, " ", -1);
+	matches = g_strsplit(nicks, " ", -1);
 	for (match = matches; *match != NULL; match++) {
 		if (strchr(*match, '*') == NULL &&
 		    strchr(*match, '?') == NULL) {
 			/* no wildcards */
                         g_string_sprintfa(str, "%s ", *match);
 		} else {
-                        get_wildcard_nicks(str, *match, channel, op, voice);
+			count = get_wildcard_nicks(str, *match, channel,
+						   op, voice);
+                        max_modes = settings_get_int("max_wildcard_modes");
+			if (max_modes > 0 && count > max_modes &&
+			    g_hash_table_lookup(optlist, "yes") == NULL) {
+                                /* too many matches */
+				g_string_free(str, TRUE);
+				cmd_params_free(free_arg);
+
+				signal_emit("error command", 1,
+					    GINT_TO_POINTER(CMDERR_NOT_GOOD_IDEA));
+				signal_stop();
+                                return NULL;
+			}
 		}
 	}
 
         if (str->len > 0) g_string_truncate(str, str->len-1);
 	ret = str->str;
 	g_string_free(str, FALSE);
+
+	cmd_params_free(free_arg);
+
+	*ret_channel = channel;
 	return ret;
 }
 
 /* SYNTAX: OP <nicks> */
 static void cmd_op(const char *data, IRC_SERVER_REC *server,
-		   IRC_CHANNEL_REC *channel)
+		   WI_ITEM_REC *item)
 {
+        IRC_CHANNEL_REC *channel;
 	char *nicks;
 
-	if (!IS_IRC_CHANNEL(channel))
-		return;
+	CMD_IRC_SERVER(server);
 
-	nicks = get_nicks(channel, data, 0, -1);
+	nicks = get_nicks(server, item, data, 0, -1, &channel);
 	if (nicks != NULL && *nicks != '\0')
-		channel_set_singlemode(server, channel->name, nicks, "+o");
+		channel_set_singlemode(channel, nicks, "+o");
 	g_free_not_null(nicks);
 }
 
 /* SYNTAX: DEOP <nicks> */
 static void cmd_deop(const char *data, IRC_SERVER_REC *server,
-		     IRC_CHANNEL_REC *channel)
+		     WI_ITEM_REC *item)
 {
+        IRC_CHANNEL_REC *channel;
 	char *nicks;
 
-	if (!IS_IRC_CHANNEL(channel))
-		return;
+        CMD_IRC_SERVER(server);
 
-	nicks = get_nicks(channel, data, 1, -1);
+	nicks = get_nicks(server, item, data, 1, -1, &channel);
 	if (nicks != NULL && *nicks != '\0')
-		channel_set_singlemode(server, channel->name, nicks, "-o");
+		channel_set_singlemode(channel, nicks, "-o");
 	g_free_not_null(nicks);
 }
 
 /* SYNTAX: VOICE <nicks> */
 static void cmd_voice(const char *data, IRC_SERVER_REC *server,
-		      IRC_CHANNEL_REC *channel)
+		      WI_ITEM_REC *item)
 {
+        IRC_CHANNEL_REC *channel;
 	char *nicks;
 
-	if (!IS_IRC_CHANNEL(channel))
-		return;
+        CMD_IRC_SERVER(server);
 
-	nicks = get_nicks(channel, data, 0, 0);
+	nicks = get_nicks(server, item, data, 0, 0, &channel);
 	if (nicks != NULL && *nicks != '\0')
-		channel_set_singlemode(server, channel->name, nicks, "+v");
+		channel_set_singlemode(channel, nicks, "+v");
 	g_free_not_null(nicks);
 }
 
 /* SYNTAX: DEVOICE <nicks> */
 static void cmd_devoice(const char *data, IRC_SERVER_REC *server,
-			IRC_CHANNEL_REC *channel)
+			WI_ITEM_REC *item)
 {
+        IRC_CHANNEL_REC *channel;
 	char *nicks;
 
-	if (!IS_IRC_CHANNEL(channel))
-		return;
+        CMD_IRC_SERVER(server);
 
-	nicks = get_nicks(channel, data, 0, 1);
+	nicks = get_nicks(server, item, data, -1, 1, &channel);
 	if (nicks != NULL && *nicks != '\0')
-		channel_set_singlemode(server, channel->name, nicks, "-v");
+		channel_set_singlemode(channel, nicks, "-v");
 	g_free_not_null(nicks);
 }
 
@@ -638,9 +713,7 @@ static void cmd_mode(const char *data, IRC_SERVER_REC *server,
 	char *target, *mode;
 	void *free_arg;
 
-	g_return_if_fail(data != NULL);
-	if (server == NULL || !server->connected || !IS_IRC_SERVER(server))
-		cmd_return_error(CMDERR_NOT_CONNECTED);
+        CMD_IRC_SERVER(server);
 
 	if (*data == '+' || *data == '-') {
 		target = "*";
@@ -663,8 +736,14 @@ static void cmd_mode(const char *data, IRC_SERVER_REC *server,
 		irc_send_cmdv(server, "MODE %s", target);
 	else if (ischannel(*target))
 		channel_set_mode(server, target, mode);
-	else
+	else {
+		if (g_strcasecmp(target, server->nick) == 0) {
+			server_redirect_event(server, "mode user", 1, target, -1, NULL,
+					      "event mode", "requested usermode change", NULL);
+		}
+
 		irc_send_cmdv(server, "MODE %s %s", target, mode);
+	}
 
 	cmd_params_free(free_arg);
 }
@@ -672,18 +751,22 @@ static void cmd_mode(const char *data, IRC_SERVER_REC *server,
 void modes_init(void)
 {
 	settings_add_str("misc", "opermode", "");
+	settings_add_int("misc", "max_wildcard_modes", 6);
 
 	signal_add("event 221", (SIGNAL_FUNC) event_user_mode);
 	signal_add("event 305", (SIGNAL_FUNC) event_unaway);
 	signal_add("event 306", (SIGNAL_FUNC) event_away);
 	signal_add("event 381", (SIGNAL_FUNC) event_oper);
 	signal_add("event mode", (SIGNAL_FUNC) event_mode);
+        signal_add("requested usermode change", (SIGNAL_FUNC) sig_req_usermode_change);
 
-	command_bind("op", NULL, (SIGNAL_FUNC) cmd_op);
-	command_bind("deop", NULL, (SIGNAL_FUNC) cmd_deop);
-	command_bind("voice", NULL, (SIGNAL_FUNC) cmd_voice);
-	command_bind("devoice", NULL, (SIGNAL_FUNC) cmd_devoice);
-	command_bind("mode", NULL, (SIGNAL_FUNC) cmd_mode);
+	command_bind_irc("op", NULL, (SIGNAL_FUNC) cmd_op);
+	command_bind_irc("deop", NULL, (SIGNAL_FUNC) cmd_deop);
+	command_bind_irc("voice", NULL, (SIGNAL_FUNC) cmd_voice);
+	command_bind_irc("devoice", NULL, (SIGNAL_FUNC) cmd_devoice);
+	command_bind_irc("mode", NULL, (SIGNAL_FUNC) cmd_mode);
+
+	command_set_options("op", "yes");
 }
 
 void modes_deinit(void)
@@ -693,6 +776,7 @@ void modes_deinit(void)
 	signal_remove("event 306", (SIGNAL_FUNC) event_away);
 	signal_remove("event 381", (SIGNAL_FUNC) event_oper);
 	signal_remove("event mode", (SIGNAL_FUNC) event_mode);
+        signal_remove("requested usermode change", (SIGNAL_FUNC) sig_req_usermode_change);
 
 	command_unbind("op", (SIGNAL_FUNC) cmd_op);
 	command_unbind("deop", (SIGNAL_FUNC) cmd_deop);
